@@ -13,6 +13,7 @@
 #include <qpdf/QPDF_Dictionary.hh>
 #include <qpdf/QPDF_Stream.hh>
 #include <qpdf/QPDF_Reserved.hh>
+#include <qpdf/Pl_Buffer.hh>
 #include <qpdf/BufferInputSource.hh>
 #include <qpdf/QPDFExc.hh>
 
@@ -481,10 +482,10 @@ QPDFObjectHandle::replaceDict(QPDFObjectHandle new_dict)
 }
 
 PointerHolder<Buffer>
-QPDFObjectHandle::getStreamData()
+QPDFObjectHandle::getStreamData(qpdf_stream_decode_level_e level)
 {
     assertStream();
-    return dynamic_cast<QPDF_Stream*>(obj.getPointer())->getStreamData();
+    return dynamic_cast<QPDF_Stream*>(obj.getPointer())->getStreamData(level);
 }
 
 PointerHolder<Buffer>
@@ -495,12 +496,45 @@ QPDFObjectHandle::getRawStreamData()
 }
 
 bool
-QPDFObjectHandle::pipeStreamData(Pipeline* p, bool filter,
-				 bool normalize, bool compress)
+QPDFObjectHandle::pipeStreamData(Pipeline* p,
+                                 unsigned long encode_flags,
+                                 qpdf_stream_decode_level_e decode_level,
+                                 bool suppress_warnings)
+{
+    return pipeStreamData(
+        p, encode_flags, decode_level, suppress_warnings, false);
+}
+
+bool
+QPDFObjectHandle::pipeStreamData(Pipeline* p,
+                                 unsigned long encode_flags,
+                                 qpdf_stream_decode_level_e decode_level,
+                                 bool suppress_warnings, bool will_retry)
 {
     assertStream();
     return dynamic_cast<QPDF_Stream*>(obj.getPointer())->pipeStreamData(
-	p, filter, normalize, compress);
+	p, encode_flags, decode_level, suppress_warnings, will_retry);
+}
+
+bool
+QPDFObjectHandle::pipeStreamData(Pipeline* p, bool filter,
+				 bool normalize, bool compress)
+{
+    unsigned long encode_flags = 0;
+    qpdf_stream_decode_level_e decode_level = qpdf_dl_none;
+    if (filter)
+    {
+        decode_level = qpdf_dl_generalized;
+        if (normalize)
+        {
+            encode_flags |= qpdf_ef_normalize;
+        }
+        if (compress)
+        {
+            encode_flags |= qpdf_ef_compress;
+        }
+    }
+    return pipeStreamData(p, encode_flags, decode_level, false);
 }
 
 void
@@ -671,6 +705,62 @@ QPDFObjectHandle::addPageContents(QPDFObjectHandle new_contents, bool first)
     this->replaceKey("/Contents", contents);
 }
 
+void
+QPDFObjectHandle::rotatePage(int angle, bool relative)
+{
+    assertPageObject();
+    if ((angle % 90) != 0)
+    {
+        throw std::runtime_error(
+            "QPDF::rotatePage called with an"
+            " angle that is not a multiple of 90");
+    }
+    int new_angle = angle;
+    if (relative)
+    {
+        int old_angle = 0;
+        bool found_rotate = false;
+        QPDFObjectHandle cur_obj = *this;
+        bool searched_parent = false;
+        std::set<QPDFObjGen> visited;
+        while (! found_rotate)
+        {
+            if (visited.count(cur_obj.getObjGen()))
+            {
+                // Don't get stuck in an infinite loop
+                break;
+            }
+            if (! visited.empty())
+            {
+                searched_parent = true;
+            }
+            visited.insert(cur_obj.getObjGen());
+            if (cur_obj.getKey("/Rotate").isInteger())
+            {
+                found_rotate = true;
+                old_angle = cur_obj.getKey("/Rotate").getIntValue();
+            }
+            else if (cur_obj.getKey("/Parent").isDictionary())
+            {
+                cur_obj = cur_obj.getKey("/Parent");
+            }
+            else
+            {
+                break;
+            }
+        }
+        QTC::TC("qpdf", "QPDFObjectHandle found old angle",
+                searched_parent ? 0 : 1);
+        if ((old_angle % 90) != 0)
+        {
+            old_angle = 0;
+        }
+        new_angle += old_angle;
+    }
+    new_angle = (new_angle + 360) % 360;
+    replaceKey("/Rotate", QPDFObjectHandle::newInteger(new_angle));
+}
+
 std::string
 QPDFObjectHandle::unparse()
 {
@@ -738,37 +828,63 @@ QPDFObjectHandle::parseContentStream(QPDFObjectHandle stream_or_array,
     {
         streams.push_back(stream_or_array);
     }
+    Pl_Buffer buf("concatenated stream data buffer");
+    std::string all_description = "content stream objects";
+    bool first = true;
     for (std::vector<QPDFObjectHandle>::iterator iter = streams.begin();
          iter != streams.end(); ++iter)
     {
         QPDFObjectHandle stream = *iter;
         if (! stream.isStream())
         {
-            throw std::logic_error(
-                "QPDFObjectHandle: parseContentStream called on non-stream");
+            QTC::TC("qpdf", "QPDFObjectHandle non-stream in parsecontent");
+            warn(stream.getOwningQPDF(),
+                 QPDFExc(qpdf_e_damaged_pdf, "content stream",
+                         "", 0,
+                         "ignoring non-stream while parsing content streams"));
         }
-        try
+        else
         {
-            parseContentStream_internal(stream, callbacks);
-        }
-        catch (TerminateParsing&)
-        {
-            return;
+            std::string og = QUtil::int_to_string(stream.getObjectID()) + " " +
+                QUtil::int_to_string(stream.getGeneration());
+            std::string description = "content stream object " + og;
+            if (first)
+            {
+                first = false;
+            }
+            else
+            {
+                all_description += ",";
+            }
+            all_description += " " + og;
+            if (! stream.pipeStreamData(&buf, 0, qpdf_dl_specialized))
+            {
+                QTC::TC("qpdf", "QPDFObjectHandle errors in parsecontent");
+                warn(stream.getOwningQPDF(),
+                     QPDFExc(qpdf_e_damaged_pdf, "content stream",
+                             description, 0,
+                             "errors while decoding content stream"));
+            }
         }
     }
+    PointerHolder<Buffer> stream_data = buf.getBuffer();
+    try
+    {
+        parseContentStream_internal(stream_data, all_description, callbacks);
+    }
+    catch (TerminateParsing&)
+    {
+        return;
+    }
     callbacks->handleEOF();
 }
 
 void
-QPDFObjectHandle::parseContentStream_internal(QPDFObjectHandle stream,
+QPDFObjectHandle::parseContentStream_internal(PointerHolder<Buffer> stream_data,
+                                              std::string const& description,
                                               ParserCallbacks* callbacks)
 {
-    stream.assertStream();
-    PointerHolder<Buffer> stream_data = stream.getStreamData();
     size_t length = stream_data->getSize();
-    std::string description = "content stream object " +
-        QUtil::int_to_string(stream.getObjectID()) + " " +
-        QUtil::int_to_string(stream.getGeneration());
     PointerHolder<InputSource> input =
         new BufferInputSource(description, stream_data.getPointer());
     QPDFTokenizer tokenizer;
@@ -777,8 +893,7 @@ QPDFObjectHandle::parseContentStream_internal(QPDFObjectHandle stream,
     while (static_cast<size_t>(input->tell()) < length)
     {
         QPDFObjectHandle obj =
-            parseInternal(input, "content", tokenizer, empty,
-                          0, 0, false, false, true);
+            parseInternal(input, "content", tokenizer, empty, 0, 0, true);
         if (! obj.isInitialized())
         {
             // EOF
@@ -839,7 +954,7 @@ QPDFObjectHandle::parse(PointerHolder<InputSource> input,
                         StringDecrypter* decrypter, QPDF* context)
 {
     return parseInternal(input, object_description, tokenizer, empty,
-                         decrypter, context, false, false, false);
+                         decrypter, context, false);
 }
 
 QPDFObjectHandle
@@ -847,26 +962,30 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
                                 std::string const& object_description,
                                 QPDFTokenizer& tokenizer, bool& empty,
                                 StringDecrypter* decrypter, QPDF* context,
-                                bool in_array, bool in_dictionary,
                                 bool content_stream)
 {
+    // This method must take care not to resolve any objects. Don't
+    // check the type of any object without first ensuring that it is
+    // a direct object. Otherwise, doing so may have the side effect
+    // of reading the object and changing the file pointer.
+
     empty = false;
-    if (in_dictionary && in_array)
-    {
-	// Although dictionaries and arrays arbitrarily nest, these
-	// variables indicate what is at the top of the stack right
-	// now, so they can, by definition, never both be true.
-	throw std::logic_error(
-	    "INTERNAL ERROR: parseInternal: in_dict && in_array");
-    }
 
     QPDFObjectHandle object;
 
-    qpdf_offset_t offset = input->tell();
-    std::vector<QPDFObjectHandle> olist;
+    std::vector<std::vector<QPDFObjectHandle> > olist_stack;
+    olist_stack.push_back(std::vector<QPDFObjectHandle>());
+    std::vector<parser_state_e> state_stack;
+    state_stack.push_back(st_top);
+    std::vector<qpdf_offset_t> offset_stack;
+    offset_stack.push_back(input->tell());
     bool done = false;
     while (! done)
     {
+        std::vector<QPDFObjectHandle>& olist = olist_stack.back();
+        parser_state_e state = state_stack.back();
+        qpdf_offset_t offset = offset_stack.back();
+
 	object = QPDFObjectHandle();
 
 	QPDFTokenizer::Token token =
@@ -877,8 +996,7 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
           case QPDFTokenizer::tt_eof:
             if (content_stream)
             {
-                // Return uninitialized object to indicate EOF
-                return object;
+                state = st_eof;
             }
             else
             {
@@ -891,54 +1009,57 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 
 	  case QPDFTokenizer::tt_brace_open:
 	  case QPDFTokenizer::tt_brace_close:
-	    // Don't know what to do with these for now
 	    QTC::TC("qpdf", "QPDFObjectHandle bad brace");
-	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-			  object_description,
-			  input->getLastOffset(),
-			  "unexpected brace token");
+            warn(context,
+                 QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                         object_description,
+                         input->getLastOffset(),
+                         "treating unexpected brace token as null"));
+            object = newNull();
 	    break;
 
 	  case QPDFTokenizer::tt_array_close:
-	    if (in_array)
+	    if (state == st_array)
 	    {
-		done = true;
+                state = st_stop;
 	    }
 	    else
 	    {
 		QTC::TC("qpdf", "QPDFObjectHandle bad array close");
-		throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-			      object_description,
-			      input->getLastOffset(),
-			      "unexpected array close token");
+                warn(context,
+                     QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                             object_description,
+                             input->getLastOffset(),
+                             "treating unexpected array close token as null"));
+                object = newNull();
 	    }
 	    break;
 
 	  case QPDFTokenizer::tt_dict_close:
-	    if (in_dictionary)
+	    if (state == st_dictionary)
 	    {
-		done = true;
+                state = st_stop;
 	    }
 	    else
 	    {
 		QTC::TC("qpdf", "QPDFObjectHandle bad dictionary close");
-		throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-			      object_description,
-			      input->getLastOffset(),
-			      "unexpected dictionary close token");
+                warn(context,
+                     QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                             object_description,
+                             input->getLastOffset(),
+                             "unexpected dictionary close token"));
+                object = newNull();
 	    }
 	    break;
 
 	  case QPDFTokenizer::tt_array_open:
-	    object = parseInternal(
-		input, object_description, tokenizer, empty,
-                decrypter, context, true, false, content_stream);
-	    break;
-
 	  case QPDFTokenizer::tt_dict_open:
-	    object = parseInternal(
-		input, object_description, tokenizer, empty,
-                decrypter, context, false, true, content_stream);
+            olist_stack.push_back(std::vector<QPDFObjectHandle>());
+            state = st_start;
+            offset_stack.push_back(input->tell());
+            state_stack.push_back(
+                (token.getType() == QPDFTokenizer::tt_array_open) ?
+                st_array : st_dictionary);
 	    break;
 
 	  case QPDFTokenizer::tt_bool:
@@ -964,12 +1085,16 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 	  case QPDFTokenizer::tt_word:
 	    {
 		std::string const& value = token.getValue();
-		if ((value == "R") && (in_array || in_dictionary) &&
-		    (olist.size() >= 2) &&
-                    (! olist.at(olist.size() - 1).isIndirect()) &&
-		    (olist.at(olist.size() - 1).isInteger()) &&
-                    (! olist.at(olist.size() - 2).isIndirect()) &&
-		    (olist.at(olist.size() - 2).isInteger()))
+                if (content_stream)
+                {
+                    object = QPDFObjectHandle::newOperator(value);
+                }
+		else if ((value == "R") && (state != st_top) &&
+                         (olist.size() >= 2) &&
+                         (! olist.at(olist.size() - 1).isIndirect()) &&
+                         (olist.at(olist.size() - 1).isInteger()) &&
+                         (! olist.at(olist.size() - 2).isIndirect()) &&
+                         (olist.at(olist.size() - 2).isInteger()))
 		{
                     if (context == 0)
                     {
@@ -986,8 +1111,7 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 		    olist.pop_back();
 		    olist.pop_back();
 		}
-		else if ((value == "endobj") &&
-			 (! (in_array || in_dictionary)))
+		else if ((value == "endobj") && (state == st_top))
 		{
 		    // We just saw endobj without having read
 		    // anything.  Treat this as a null and do not move
@@ -996,17 +1120,16 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 		    input->seek(input->getLastOffset(), SEEK_SET);
                     empty = true;
 		}
-		else if (content_stream)
-                {
-                    object = QPDFObjectHandle::newOperator(token.getValue());
-                }
 		else
 		{
-		    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-				  object_description,
-				  input->getLastOffset(),
-				  "unknown token while reading object (" +
-				  value + ")");
+                    QTC::TC("qpdf", "QPDFObjectHandle treat word as string");
+                    warn(context,
+                         QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                                 object_description,
+                                 input->getLastOffset(),
+                                 "unknown token while reading object;"
+                                 " treating as string"));
+                    object = newString(value);
 		}
 	    }
 	    break;
@@ -1024,64 +1147,142 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 	    break;
 
 	  default:
-	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-			  object_description,
-			  input->getLastOffset(),
-			  "unknown token type while reading object");
+            warn(context,
+                 QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                         object_description,
+                         input->getLastOffset(),
+                         "treating unknown token type as null while "
+                         "reading object"));
+            object = newNull();
 	    break;
 	}
 
-	if (in_dictionary || in_array)
-	{
-	    if (! done)
-	    {
-		olist.push_back(object);
-	    }
-	}
-	else if (! object.isInitialized())
-	{
-	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
-			  object_description,
-			  input->getLastOffset(),
-			  "parse error while reading object");
-	}
-	else
-	{
-	    done = true;
-	}
-    }
+        if ((! object.isInitialized()) &&
+            (! ((state == st_start) ||
+                (state == st_stop) ||
+                (state == st_eof))))
+        {
+            throw std::logic_error(
+                "QPDFObjectHandle::parseInternal: "
+                "unexpected uninitialized object");
+            object = newNull();
+        }
 
-    if (in_array)
-    {
-	object = newArray(olist);
-    }
-    else if (in_dictionary)
-    {
-	// Convert list to map.  Alternating elements are keys.
-	std::map<std::string, QPDFObjectHandle> dict;
-	if (olist.size() % 2)
-	{
-	    QTC::TC("qpdf", "QPDFObjectHandle dictionary odd number of elements");
-	    throw QPDFExc(
-		qpdf_e_damaged_pdf, input->getName(),
-		object_description, input->getLastOffset(),
-		"dictionary ending here has an odd number of elements");
-	}
-	for (unsigned int i = 0; i < olist.size(); i += 2)
-	{
-	    QPDFObjectHandle key_obj = olist.at(i);
-	    QPDFObjectHandle val = olist.at(i + 1);
-	    if (! key_obj.isName())
-	    {
-		throw QPDFExc(
-		    qpdf_e_damaged_pdf,
-		    input->getName(), object_description, offset,
-		    std::string("dictionary key not name (") +
-		    key_obj.unparse() + ")");
-	    }
-	    dict[key_obj.getName()] = val;
-	}
-	object = newDictionary(dict);
+        switch (state)
+        {
+          case st_eof:
+            if (state_stack.size() > 1)
+            {
+                warn(context,
+                     QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+                             object_description,
+                             input->getLastOffset(),
+                             "parse error while reading object"));
+            }
+            done = true;
+            // Leave object uninitialized to indicate EOF
+            break;
+
+          case st_dictionary:
+          case st_array:
+            olist.push_back(object);
+            break;
+
+          case st_top:
+            done = true;
+            break;
+
+          case st_start:
+            break;
+
+          case st_stop:
+            if ((state_stack.size() < 2) || (olist_stack.size() < 2))
+            {
+                throw std::logic_error(
+                    "QPDFObjectHandle::parseInternal: st_stop encountered"
+                    " with insufficient elements in stack");
+            }
+            parser_state_e old_state = state_stack.back();
+            state_stack.pop_back();
+            if (old_state == st_array)
+            {
+                object = newArray(olist);
+            }
+            else if (old_state == st_dictionary)
+            {
+                // Convert list to map. Alternating elements are keys.
+                // Attempt to recover more or less gracefully from
+                // invalid dictionaries.
+                std::set<std::string> names;
+                for (std::vector<QPDFObjectHandle>::iterator iter =
+                         olist.begin();
+                     iter != olist.end(); ++iter)
+                {
+                    if ((! (*iter).isIndirect()) && (*iter).isName())
+                    {
+                        names.insert((*iter).getName());
+                    }
+                }
+
+                std::map<std::string, QPDFObjectHandle> dict;
+                int next_fake_key = 1;
+                for (unsigned int i = 0; i < olist.size(); ++i)
+                {
+                    QPDFObjectHandle key_obj = olist.at(i);
+                    QPDFObjectHandle val;
+                    if (key_obj.isIndirect() || (! key_obj.isName()))
+                    {
+                        bool found_fake = false;
+                        std::string candidate;
+                        while (! found_fake)
+                        {
+                            candidate =
+                                "/QPDFFake" +
+                                QUtil::int_to_string(next_fake_key++);
+                            found_fake = (names.count(candidate) == 0);
+                            QTC::TC("qpdf", "QPDFObjectHandle found fake",
+                                    (found_fake ? 0 : 1));
+                        }
+                        warn(context,
+                             QPDFExc(
+                                 qpdf_e_damaged_pdf,
+                                 input->getName(), object_description, offset,
+                                 "expected dictionary key but found"
+                                 " non-name object; inserting key " +
+                                 candidate));
+                        val = key_obj;
+                        key_obj = newName(candidate);
+                    }
+                    else if (i + 1 >= olist.size())
+                    {
+                        QTC::TC("qpdf", "QPDFObjectHandle no val for last key");
+                        warn(context,
+                             QPDFExc(
+                                 qpdf_e_damaged_pdf,
+                                 input->getName(), object_description, offset,
+                                 "dictionary ended prematurely; "
+                                 "using null as value for last key"));
+                        val = newNull();
+                    }
+                    else
+                    {
+                        val = olist.at(++i);
+                    }
+                    dict[key_obj.getName()] = val;
+                }
+                object = newDictionary(dict);
+            }
+            olist_stack.pop_back();
+            offset_stack.pop_back();
+            if (state_stack.back() == st_top)
+            {
+                done = true;
+            }
+            else
+            {
+                olist_stack.back().push_back(object);
+            }
+        }
     }
 
     return object;
@@ -1090,6 +1291,16 @@ QPDFObjectHandle::parseInternal(PointerHolder<InputSource> input,
 QPDFObjectHandle
 QPDFObjectHandle::newIndirect(QPDF* qpdf, int objid, int generation)
 {
+    if (objid == 0)
+    {
+        // Special case: QPDF uses objid 0 as a sentinel for direct
+        // objects, and the PDF specification doesn't allow for object
+        // 0. Treat indirect references to object 0 as null so that we
+        // never create an indirect object with objid 0.
+        QTC::TC("qpdf", "QPDFObjectHandle indirect with 0 objid");
+        return newNull();
+    }
+
     return QPDFObjectHandle(qpdf, objid, generation);
 }
 
@@ -1535,3 +1746,20 @@ QPDFObjectHandle::dereference()
         }
     }
 }
+
+void
+QPDFObjectHandle::warn(QPDF* qpdf, QPDFExc const& e)
+{
+    // If parsing on behalf of a QPDF object and want to give a
+    // warning, we can warn through the object. If parsing for some
+    // other reason, such as an explicit creation of an object from a
+    // string, then just throw the exception.
+    if (qpdf)
+    {
+        QPDF::Warner::warn(qpdf, e);
+    }
+    else
+    {
+        throw e;
+    }
+}
